/**************************************************************************************
 * Copyright (C) 2006 Thomas Bernhardt. All rights reserved.                          *
 * http://esper.codehaus.org                                                          *
 * ---------------------------------------------------------------------------------- *
 * The software in this package is published under the terms of the GPL license       *
 * a copy of which has been included with this distribution in the license.txt file.  *
 **************************************************************************************/
package net.esper.client;

import java.util.List;
import java.util.LinkedList;

/**
 * Configuration information for legacy (non-JavaBean) event types.
 */
public class ConfigurationEventTypeLegacy
{
    private AccessorStyle accessorStyle;
    private CodeGeneration codeGeneration;
    private List<LegacyMethodPropDesc> methodProperties;
    private List<LegacyFieldPropDesc> fieldProperties;

    /**
     * Ctor.
     */
    public ConfigurationEventTypeLegacy()
    {
        accessorStyle = AccessorStyle.JAVABEAN;
        codeGeneration = CodeGeneration.ENABLED;
        methodProperties = new LinkedList<LegacyMethodPropDesc>();
        fieldProperties = new LinkedList<LegacyFieldPropDesc>();
    }

    /**
     * Accessor style defines the methods of a class that are automatically exposed via event property.
     */
    public enum AccessorStyle {     // ensure the names match the configuration schema type restriction defs
        /**
         * Expose JavaBean-style getter methods only, plus explicitly configured properties.
         */
        JAVABEAN,

        /**
         * Expose only the explicitly configured methods and public members as event properties.
         */
        EXPLICIT,

        /**
         * Expose all public methods and public members as event properties, plus explicitly configured properties.
         */
        PUBLIC
    }

    /**
     * Enum to control code generation.
     */
    public enum CodeGeneration {    // ensure the names match the configuration schema type restriction defs
        /**
         * Enables code generation.
         */
        ENABLED,

        /**
         * Dispables code generation.
         */
        DISABLED
    }

    /**
     * Sets the accessor style. Thus controls how the engine exposes event properties
     * based on a Java class's public methods and public member variables.
     * @param accessorStyle is the style enum
     */
    public void setAccessorStyle(AccessorStyle accessorStyle)
    {
        this.accessorStyle = accessorStyle;
    }

    /**
     * Sets the code generation flags. Thus controls whether or not the engine
     * generates code for access to event property values.
     * @param codeGeneration is the code generation enum
     */
    public void setCodeGeneration(CodeGeneration codeGeneration)
    {
        this.codeGeneration = codeGeneration;
    }

    /**
     * Returns the accessor style.
     * @return accessor style
     */
    public AccessorStyle getAccessorStyle()
    {
        return accessorStyle;
    }

    /**
     * Returns the code generation flag.
     * @return code generation flag
     */
    public CodeGeneration getCodeGeneration()
    {
        return codeGeneration;
    }

    /**
     * Returns a list of descriptors specifying explicitly configured method names
     * and their property name.
     * @return list of explicit method-access descriptors
     */
    public List<LegacyMethodPropDesc> getMethodProperties()
    {
        return methodProperties;
    }

    /**
     * Returns a list of descriptors specifying explicitly configured field names
     * and their property name.
     * @return list of explicit field-access descriptors
     */
    public List<LegacyFieldPropDesc> getFieldProperties()
    {
        return fieldProperties;
    }

    /**
     * Adds the named event property backed by the named accessor method.
     * <p>
     * The accessor method is expected to be a public method with no parameters
     * for simple event properties, or with a single integer parameter for indexed
     * event properties, or with a single String parameter for mapped event properties.
     * @param name is the event property name
     * @param accessorMethod is the accessor method name.
     */
    public void addMethodProperty(String name, String accessorMethod)
    {
        methodProperties.add(new LegacyMethodPropDesc(name, accessorMethod));
    }

    /**
     * Adds the named event property backed by the named accessor field.
     * @param name is the event property name
     * @param accessorField is the accessor field underlying the name
     */
    public void addFieldProperty(String name, String accessorField)
    {
        fieldProperties.add(new LegacyFieldPropDesc(name, accessorField));
    }

    /**
     * Encapsulates information about an accessor field backing a named event property.
     */
    public class LegacyFieldPropDesc
    {
        private String name;
        private String accessorFieldName;

        /**
         * Ctor.
         * @param name is the event property name
         * @param accessorFieldName is the accessor field name
         */
        public LegacyFieldPropDesc(String name, String accessorFieldName)
        {
            this.name = name;
            this.accessorFieldName = accessorFieldName;
        }

        /**
         * Returns the event property name.
         * @return event property name
         */
        public String getName()
        {
            return name;
        }

        /**
         * Returns the accessor field name.
         * @return accessor field name
         */
        public String getAccessorFieldName()
        {
            return accessorFieldName;
        }
    }

    /**
     * Encapsulates information about an accessor method backing a named event property.
     */
    public class LegacyMethodPropDesc
    {
        private String name;
        private String accessorMethodName;

        /**
         * Ctor.
         * @param name is the event property name
         * @param accessorMethodName is the name of the accessor method
         */
        public LegacyMethodPropDesc(String name, String accessorMethodName)
        {
            this.name = name;
            this.accessorMethodName = accessorMethodName;
        }

        /**
         * Returns the event property name.
         * @return event property name
         */
        public String getName()
        {
            return name;
        }

        /**
         * Returns the accessor method name.
         * @return accessor method name
         */
        public String getAccessorMethodName()
        {
            return accessorMethodName;
        }
    }
}
